@@ -14,7 +14,7 @@ from . import ctx as mitmproxy_ctx
 
 class Master:
     """
-        The master handles mitmproxy's main event loop.
+    The master handles mitmproxy's main event loop.
     """
 
     event_loop: asyncio.AbstractEventLoop
@@ -46,9 +46,11 @@ class Master:
             # Handle scheduled tasks (configure()) first.
             await asyncio.sleep(0)
             await self.running()
-            await self.should_exit.wait()
-
-            await self.done()
+            try:
+                await self.should_exit.wait()
+            finally:
+                # .wait might be cancelled (e.g. by sys.exit)
+                await self.done()
         finally:
             self.event_loop.set_exception_handler(old_handler)
 
@@ -66,18 +68,22 @@ class Master:
         await self.addons.trigger_event(hooks.DoneHook())
 
     def _asyncio_exception_handler(self, loop, context):
-        exc: Exception = context["exception"]
-        if isinstance(exc, OSError) and exc.errno == 10038:
-            return  # suppress https://bugs.python.org/issue43253
-        self.log.error(
-            "\n".join(traceback.format_exception(
-                type(exc),
-                exc,
-                exc.__traceback__
-            )) +
-            "\nPlease lodge a bug report at:" +
-            "\n\thttps://github.com/mitmproxy/mitmproxy/issues"
-        )
+        try:
+            exc: Exception = context["exception"]
+        except KeyError:
+            self.log.error(
+                f"Unhandled asyncio error: {context}"
+                "\nPlease lodge a bug report at:"
+                + "\n\thttps://github.com/mitmproxy/mitmproxy/issues"
+            )
+        else:
+            if isinstance(exc, OSError) and exc.errno == 10038:
+                return  # suppress https://bugs.python.org/issue43253
+            self.log.error(
+                "\n".join(traceback.format_exception(type(exc), exc, exc.__traceback__))
+                + "\nPlease lodge a bug report at:"
+                + "\n\thttps://github.com/mitmproxy/mitmproxy/issues"
+            )
 
     async def load_flow(self, f):
         """
