@@ -1,5 +1,5 @@
 import abc
-from typing import Callable, Optional, Type, Union
+from typing import Callable, Optional, Union
 
 import h11
 from h11._readers import ChunkedReader, ContentLengthReader, Http10Reader
@@ -13,8 +13,17 @@ from mitmproxy.proxy.layers.http._base import ReceiveHttp, StreamId
 from mitmproxy.proxy.utils import expect
 from mitmproxy.utils import human
 from ._base import HttpConnection, format_error
-from ._events import HttpEvent, RequestData, RequestEndOfMessage, RequestHeaders, RequestProtocolError, ResponseData, \
-    ResponseEndOfMessage, ResponseHeaders, ResponseProtocolError
+from ._events import (
+    HttpEvent,
+    RequestData,
+    RequestEndOfMessage,
+    RequestHeaders,
+    RequestProtocolError,
+    ResponseData,
+    ResponseEndOfMessage,
+    ResponseHeaders,
+    ResponseProtocolError,
+)
 from ...context import Context
 
 TBodyReader = Union[ChunkedReader, Http10Reader, ContentLengthReader]
@@ -31,9 +40,9 @@ class Http1Connection(HttpConnection, metaclass=abc.ABCMeta):
     body_reader: TBodyReader
     buf: ReceiveBuffer
 
-    ReceiveProtocolError: Type[Union[RequestProtocolError, ResponseProtocolError]]
-    ReceiveData: Type[Union[RequestData, ResponseData]]
-    ReceiveEndOfMessage: Type[Union[RequestEndOfMessage, ResponseEndOfMessage]]
+    ReceiveProtocolError: type[Union[RequestProtocolError, ResponseProtocolError]]
+    ReceiveData: type[Union[RequestData, ResponseData]]
+    ReceiveEndOfMessage: type[Union[RequestEndOfMessage, ResponseEndOfMessage]]
 
     def __init__(self, context: Context, conn: Connection):
         super().__init__(context, conn)
@@ -44,14 +53,19 @@ class Http1Connection(HttpConnection, metaclass=abc.ABCMeta):
         yield from ()  # pragma: no cover
 
     @abc.abstractmethod
-    def read_headers(self, event: events.ConnectionEvent) -> layer.CommandGenerator[None]:
+    def read_headers(
+        self, event: events.ConnectionEvent
+    ) -> layer.CommandGenerator[None]:
         yield from ()  # pragma: no cover
 
     def _handle_event(self, event: events.Event) -> layer.CommandGenerator[None]:
         if isinstance(event, HttpEvent):
             yield from self.send(event)
         else:
-            if isinstance(event, events.DataReceived) and self.state != self.passthrough:
+            if (
+                isinstance(event, events.DataReceived)
+                and self.state != self.passthrough
+            ):
                 self.buf += event.data
             yield from self.state(event)
 
@@ -74,7 +88,11 @@ class Http1Connection(HttpConnection, metaclass=abc.ABCMeta):
                     raise AssertionError(f"Unexpected event: {event}")
             except h11.ProtocolError as e:
                 yield commands.CloseConnection(self.conn)
-                yield ReceiveHttp(self.ReceiveProtocolError(self.stream_id, f"HTTP/1 protocol error: {e}"))
+                yield ReceiveHttp(
+                    self.ReceiveProtocolError(
+                        self.stream_id, f"HTTP/1 protocol error: {e}"
+                    )
+                )
                 return
 
             if h11_event is None:
@@ -90,10 +108,7 @@ class Http1Connection(HttpConnection, metaclass=abc.ABCMeta):
                 if self.request.data.method.upper() != b"CONNECT":
                     yield ReceiveHttp(self.ReceiveEndOfMessage(self.stream_id))
                 is_request = isinstance(self, Http1Server)
-                yield from self.mark_done(
-                    request=is_request,
-                    response=not is_request
-                )
+                yield from self.mark_done(request=is_request, response=not is_request)
                 return
 
     def wait(self, event: events.Event) -> layer.CommandGenerator[None]:
@@ -110,8 +125,13 @@ class Http1Connection(HttpConnection, metaclass=abc.ABCMeta):
             # see https://github.com/httpwg/http-core/issues/22
             if event.connection.state is not ConnectionState.CLOSED:
                 yield commands.CloseConnection(event.connection)
-            yield ReceiveHttp(self.ReceiveProtocolError(self.stream_id, f"Client disconnected.",
-                                                        code=status_codes.CLIENT_CLOSED_REQUEST))
+            yield ReceiveHttp(
+                self.ReceiveProtocolError(
+                    self.stream_id,
+                    f"Client disconnected.",
+                    code=status_codes.CLIENT_CLOSED_REQUEST,
+                )
+            )
         else:  # pragma: no cover
             raise AssertionError(f"Unexpected event: {event}")
 
@@ -122,7 +142,10 @@ class Http1Connection(HttpConnection, metaclass=abc.ABCMeta):
         self.state = self.passthrough
         if self.buf:
             already_received = self.buf.maybe_extract_at_most(len(self.buf))
-            yield from self.state(events.DataReceived(self.conn, already_received))
+            # Some clients send superfluous newlines after CONNECT, we want to eat those.
+            already_received = already_received.lstrip(b"\r\n")
+            if already_received:
+                yield from self.state(events.DataReceived(self.conn, already_received))
 
     def passthrough(self, event: events.Event) -> layer.CommandGenerator[None]:
         assert self.stream_id
@@ -134,7 +157,9 @@ class Http1Connection(HttpConnection, metaclass=abc.ABCMeta):
             else:
                 yield ReceiveHttp(ResponseEndOfMessage(self.stream_id))
 
-    def mark_done(self, *, request: bool = False, response: bool = False) -> layer.CommandGenerator[None]:
+    def mark_done(
+        self, *, request: bool = False, response: bool = False
+    ) -> layer.CommandGenerator[None]:
         if request:
             self.request_done = True
         if response:
@@ -146,15 +171,21 @@ class Http1Connection(HttpConnection, metaclass=abc.ABCMeta):
                 yield from self.make_pipe()
                 return
             try:
-                read_until_eof_semantics = http1.expected_http_body_size(self.request, self.response) == -1
+                read_until_eof_semantics = (
+                    http1.expected_http_body_size(self.request, self.response) == -1
+                )
             except ValueError:
                 # this may raise only now (and not earlier) because an addon set invalid headers,
                 # in which case it's not really clear what we are supposed to do.
                 read_until_eof_semantics = False
             connection_done = (
                 read_until_eof_semantics
-                or http1.connection_close(self.request.http_version, self.request.headers)
-                or http1.connection_close(self.response.http_version, self.response.headers)
+                or http1.connection_close(
+                    self.request.http_version, self.request.headers
+                )
+                or http1.connection_close(
+                    self.response.http_version, self.response.headers
+                )
                 # If we proxy HTTP/2 to HTTP/1, we only use upstream connections for one request.
                 # This simplifies our connection management quite a bit as we can rely on
                 # the proxyserver's max-connection-per-server throttling.
@@ -218,27 +249,50 @@ class Http1Server(Http1Connection):
             yield from self.mark_done(response=True)
         elif isinstance(event, ResponseProtocolError):
             if not self.response and event.code != status_codes.NO_RESPONSE:
-                yield commands.SendData(self.conn, make_error_response(event.code, event.message))
+                yield commands.SendData(
+                    self.conn, make_error_response(event.code, event.message)
+                )
             if self.conn.state & ConnectionState.CAN_WRITE:
                 yield commands.CloseConnection(self.conn)
         else:
             raise AssertionError(f"Unexpected event: {event}")
 
-    def read_headers(self, event: events.ConnectionEvent) -> layer.CommandGenerator[None]:
+    def read_headers(
+        self, event: events.ConnectionEvent
+    ) -> layer.CommandGenerator[None]:
         if isinstance(event, events.DataReceived):
             request_head = self.buf.maybe_extract_lines()
             if request_head:
-                request_head = [bytes(x) for x in request_head]  # TODO: Make url.parse compatible with bytearrays
+                request_head = [
+                    bytes(x) for x in request_head
+                ]  # TODO: Make url.parse compatible with bytearrays
                 try:
                     self.request = http1.read_request_head(request_head)
+                    if self.context.options.validate_inbound_headers:
+                        http1.validate_headers(self.request.headers)
                     expected_body_size = http1.expected_http_body_size(self.request)
                 except ValueError as e:
-                    yield commands.Log(f"{human.format_address(self.conn.peername)}: {e}")
                     yield commands.SendData(self.conn, make_error_response(400, str(e)))
                     yield commands.CloseConnection(self.conn)
+                    if self.request:
+                        # we have headers that we can show in the ui
+                        yield ReceiveHttp(
+                            RequestHeaders(self.stream_id, self.request, False)
+                        )
+                        yield ReceiveHttp(
+                            RequestProtocolError(self.stream_id, str(e), 400)
+                        )
+                    else:
+                        yield commands.Log(
+                            f"{human.format_address(self.conn.peername)}: {e}"
+                        )
                     self.state = self.done
                     return
-                yield ReceiveHttp(RequestHeaders(self.stream_id, self.request, expected_body_size == 0))
+                yield ReceiveHttp(
+                    RequestHeaders(
+                        self.stream_id, self.request, expected_body_size == 0
+                    )
+                )
                 self.body_reader = make_body_reader(expected_body_size)
                 self.state = self.read_body
                 yield from self.state(event)
@@ -247,12 +301,16 @@ class Http1Server(Http1Connection):
         elif isinstance(event, events.ConnectionClosed):
             buf = bytes(self.buf)
             if buf.strip():
-                yield commands.Log(f"Client closed connection before completing request headers: {buf!r}")
+                yield commands.Log(
+                    f"Client closed connection before completing request headers: {buf!r}"
+                )
             yield commands.CloseConnection(self.conn)
         else:
             raise AssertionError(f"Unexpected event: {event}")
 
-    def mark_done(self, *, request: bool = False, response: bool = False) -> layer.CommandGenerator[None]:
+    def mark_done(
+        self, *, request: bool = False, response: bool = False
+    ) -> layer.CommandGenerator[None]:
         yield from super().mark_done(request=request, response=response)
         if self.request_done and not self.response_done:
             self.state = self.wait
@@ -283,11 +341,19 @@ class Http1Client(Http1Connection):
             request = event.request
             if request.is_http2:
                 # Convert to an HTTP/1 request.
-                request = request.copy()  # (we could probably be a bit more efficient here.)
+                request = (
+                    request.copy()
+                )  # (we could probably be a bit more efficient here.)
                 request.http_version = "HTTP/1.1"
                 if "Host" not in request.headers and request.authority:
                     request.headers.insert(0, "Host", request.authority)
                 request.authority = ""
+                cookie_headers = request.headers.get_all("Cookie")
+                if len(cookie_headers) > 1:
+                    # Only HTTP/2 supports multiple cookie headers, HTTP/1.x does not.
+                    # see: https://www.rfc-editor.org/rfc/rfc6265#section-5.4
+                    #      https://www.rfc-editor.org/rfc/rfc7540#section-8.1.2.5
+                    request.headers["Cookie"] = "; ".join(cookie_headers)
             raw = http1.assemble_request_head(request)
             yield commands.SendData(self.conn, raw)
         elif isinstance(event, RequestData):
@@ -308,7 +374,9 @@ class Http1Client(Http1Connection):
         else:
             raise AssertionError(f"Unexpected event: {event}")
 
-    def read_headers(self, event: events.ConnectionEvent) -> layer.CommandGenerator[None]:
+    def read_headers(
+        self, event: events.ConnectionEvent
+    ) -> layer.CommandGenerator[None]:
         if isinstance(event, events.DataReceived):
             if not self.request:
                 # we just received some data for an unknown request.
@@ -319,15 +387,27 @@ class Http1Client(Http1Connection):
 
             response_head = self.buf.maybe_extract_lines()
             if response_head:
-                response_head = [bytes(x) for x in response_head]  # TODO: Make url.parse compatible with bytearrays
+                response_head = [
+                    bytes(x) for x in response_head
+                ]  # TODO: Make url.parse compatible with bytearrays
                 try:
                     self.response = http1.read_response_head(response_head)
-                    expected_size = http1.expected_http_body_size(self.request, self.response)
+                    if self.context.options.validate_inbound_headers:
+                        http1.validate_headers(self.response.headers)
+                    expected_size = http1.expected_http_body_size(
+                        self.request, self.response
+                    )
                 except ValueError as e:
                     yield commands.CloseConnection(self.conn)
-                    yield ReceiveHttp(ResponseProtocolError(self.stream_id, f"Cannot parse HTTP response: {e}"))
+                    yield ReceiveHttp(
+                        ResponseProtocolError(
+                            self.stream_id, f"Cannot parse HTTP response: {e}"
+                        )
+                    )
                     return
-                yield ReceiveHttp(ResponseHeaders(self.stream_id, self.response, expected_size == 0))
+                yield ReceiveHttp(
+                    ResponseHeaders(self.stream_id, self.response, expected_size == 0)
+                )
                 self.body_reader = make_body_reader(expected_size)
 
                 self.state = self.read_body
@@ -339,13 +419,21 @@ class Http1Client(Http1Connection):
                 yield commands.CloseConnection(self.conn)
             if self.stream_id:
                 if self.buf:
-                    yield ReceiveHttp(ResponseProtocolError(self.stream_id,
-                                                            f"unexpected server response: {bytes(self.buf)!r}"))
+                    yield ReceiveHttp(
+                        ResponseProtocolError(
+                            self.stream_id,
+                            f"unexpected server response: {bytes(self.buf)!r}",
+                        )
+                    )
                 else:
                     # The server has closed the connection to prevent us from continuing.
                     # We need to signal that to the stream.
                     # https://tools.ietf.org/html/rfc7231#section-6.5.11
-                    yield ReceiveHttp(ResponseProtocolError(self.stream_id, "server closed connection"))
+                    yield ReceiveHttp(
+                        ResponseProtocolError(
+                            self.stream_id, "server closed connection"
+                        )
+                    )
             else:
                 return
         else:
@@ -381,7 +469,7 @@ def make_error_response(
             Server=version.MITMPROXY,
             Connection="close",
             Content_Type="text/html",
-        )
+        ),
     )
     return http1.assemble_response(resp)
 
