@@ -1,30 +1,20 @@
 import webbrowser
+from collections.abc import Sequence
 
 from mitmproxy import ctx
-from typing import Sequence
 
 
 class WebAddon:
     def load(self, loader):
+        loader.add_option("web_open_browser", bool, True, "Start a browser.")
+        loader.add_option("web_debug", bool, False, "Enable mitmweb debugging.")
+        loader.add_option("web_port", int, 8081, "Web UI port.")
+        loader.add_option("web_host", str, "127.0.0.1", "Web UI host.")
         loader.add_option(
-            "web_open_browser", bool, True,
-            "Start a browser."
-        )
-        loader.add_option(
-            "web_debug", bool, False,
-            "Enable mitmweb debugging."
-        )
-        loader.add_option(
-            "web_port", int, 8081,
-            "Web UI port."
-        )
-        loader.add_option(
-            "web_host", str, "127.0.0.1",
-            "Web UI host."
-        )
-        loader.add_option(
-            "web_columns", Sequence[str], ["tls", "icon", "path", "method", "status", "size", "time"],
-            "Columns to show in the flow list"
+            "web_columns",
+            Sequence[str],
+            ["tls", "icon", "path", "method", "status", "size", "time"],
+            "Columns to show in the flow list",
         )
 
     def running(self):
@@ -49,11 +39,19 @@ def open_browser(url: str) -> bool:
         False, if no suitable browser has been found.
     """
     browsers = (
-        "windows-default", "macosx",
+        "windows-default",
+        "macosx",
         "wslview %s",
-        "x-www-browser %s", "gnome-open %s", "xdg-open",
-        "google-chrome", "chrome", "chromium", "chromium-browser",
-        "firefox", "opera", "safari",
+        "x-www-browser %s",
+        "gnome-open %s",
+        "xdg-open",
+        "google-chrome",
+        "chrome",
+        "chromium",
+        "chromium-browser",
+        "firefox",
+        "opera",
+        "safari",
     )
     for browser in browsers:
         try:
